#!/usr/local/bin/python
#
"""SLB - Super Light Blog
"""
#
# Copyright (c) 2005 Satoshi Fukutomi <info@fuktommy.com>.
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions
# are met:
# 1. Redistributions of source code must retain the above copyright
#    notice, this list of conditions and the following disclaimer.
# 2. Redistributions in binary form must reproduce the above copyright
#    notice, this list of conditions and the following disclaimer in the
#    documentation and/or other materials provided with the distribution.
#
# THIS SOFTWARE IS PROVIDED BY THE REGENTS AND CONTRIBUTORS ``AS IS'' AND
# ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
# ARE DISCLAIMED.  IN NO EVENT SHALL THE REGENTS OR CONTRIBUTORS BE LIABLE
# FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
# DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
# OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
# HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
# LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
# OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
# SUCH DAMAGE.
#

import re
import os
import cgi
import md5
import time
from urlparse import urljoin
from xml.sax import parse, SAXParseException
from xml.sax.handler import ContentHandler

class Config(dict):
    def __init__(self, filename=None):
        dict.__init__(self)
        if filename:
            self.read(filename)

    def read(self, filename):
        for line in file(filename):
            key, val = line.split(":", 1)
            self[key.strip()] = val.strip()

    def getint(self, key):
        return int(self[key])

config = Config("/virtual/fuktommy/private/blog.conf")

class Item:
    "RSS item"

    def __init__(self, link="", title="", date="",
                 description="", content=""):
        self.link  = link
        self.date  = date
        self.title = cgi.escape(title)
        self.description = cgi.escape(description)
        self.content = content.replace("]]>", "&#93;&#93;>")

# End of Item

class RSS(dict):
    """RSS.

    It is the dictionary which key is URI.
    """

    encode = "utf-8"
    lang   = "en"
    title  = ""
    parent = ""         # Place where is documents or RSS
    link   = ""         # URI of main page
    uri    = ""         # URI of RSS
    description = ""

    def __init__(self, encode="utf-8", lang="ja", title="",
            parent="", link="", uri="", description="", xsl=""):
        self.encode = encode
        self.lang   = lang

        if title:
            self.title = title
        else:
            self.title = config["title"]
        if parent:
            self.parent = parent
        else:
            self.parent = config["link"]
        if link:
            self.link = link
        else:
            self.link = config["link"]
        if uri:
            self.uri = uri
        else:
            self.uri = config["rssuri"]
        if description:
            self.description = description
        else:
            self.description = config["description"]
        if xsl:
            self.xsl = xsl
        else:
            self.xsl = config["xsl"]

        if self.parent[-1] != "/":
            self.parent += "/"

    def append(self, link, title="", date="",
               description="", content=""):
        """Add an item."""
        item = Item(link, title=title, date=date,
                    description=description, content=content)
        self[link] = item

    def keys(self):
        """List of links sorted by date."""
        links = dict.keys(self)
        links.sort(lambda x,y: cmp(self[y].date, self[x].date))
        return links

    def __iter__(self):
        return iter(self.keys())

# Enf of RSS


def make_rss1(rss):
    """Generate RSS 1.0."""
    buf = ""
    buf += '<?xml version="1.0" encoding="' + rss.encode + '"?>\n'
    if rss.xsl:
        buf += '<?xml-stylesheet href="%s" type="text/xsl"?>\n' % rss.xsl
    buf += '<rdf:RDF\n'
    buf += '  xmlns="http://purl.org/rss/1.0/"\n'
    buf += '  xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#"\n'
    buf += '  xmlns:content="http://purl.org/rss/1.0/modules/content/"\n'
    buf += '  xmlns:dc="http://purl.org/dc/elements/1.1/"\n'
    buf += '  xml:lang="' + rss.lang + '">\n'
    buf += '<channel rdf:about="' + rss.uri + '">\n'
    buf += '  <title>' + rss.title + '</title>\n'
    buf += '  <link>' + rss.link + '</link>\n'
    buf += '  <description>' + rss.description + '</description>\n'
    buf += '  <items><rdf:Seq>\n'

    for f in rss:
        buf += '    <rdf:li rdf:resource="%s"/>\n' % f
    buf += '  </rdf:Seq></items>\n'
    buf += '</channel>\n'

    for uri in rss:
        f = rss[uri]
        buf += '<item rdf:about="%s">\n' % f.link
        buf += '  <title>%s</title>\n' % f.title
        buf += '  <link>%s</link>\n' % f.link
        buf += '  <dc:date>%s</dc:date>\n' % f.date
        if f.description:
              buf += '  <description>' + f.description + '</description>\n'
        if f.content:
              buf += '  <content:encoded><![CDATA[%s]]></content:encoded>\n' \
                % f.content
        buf += '</item>\n'

    buf += '</rdf:RDF>\n'
    return buf
# End of make_rss1


class RSShandler(ContentHandler):
    elements = ["title", "link", "dc:date", "description", "content:encoded"]

    def __init__(self):
        ContentHandler.__init__(self)
        self.items = []
        self.state = ""

    def startElement(self, name, attrs):
        if name == "item":
            self.items.append(Item())
            self.state = "item"
        elif (self.state == "item") and (name in self.elements):
            self.state = name

    def endElement(self, name):
        if self.state and (name in self.elements):
            self.state = "item"

    def characters(self, contents):
        contents = contents.encode("utf-8")
        if (not self.state) or (self.state == "item"):
            pass
        elif self.state == "title":
            self.items[-1].title += contents
        elif self.state == "link":
            self.items[-1].link += contents
        elif self.state == "dc:date":
            self.items[-1].date += contents
        elif self.state == "description":
            self.items[-1].description += contents
        elif self.state == "content:encoded":
            self.items[-1].content += contents

# End of RSShandler


class Entry:
    """An article."""

    def __init__(self, id):
        self.id = id

    def update(self, title="", content=""):
        if not (title or content):
            self.remove()
            return
        description = re.sub(r"<.*?>", "", content)
        description = unicode(description, "utf-8")
        description = description[:config.getint("descriptionsize")]
        description = description.encode("utf-8")
        link = urljoin(config["link"], "#e%d" % self.id)
        w3cdate = time.strftime('%Y-%m-%dT%H:%M:%S', time.localtime(self.id))
        w3cdate += config["timezone"]
        self.item = Item(link = link,
                         title = title,
                         date = w3cdate,
                         description = description,
                         content = content)
        return self.item

    def month(self):
        return time.strftime("%Y-%m", time.localtime(self.id))

    def dirname(self):
        return os.path.join(config["xmldir"], self.month())

    def path(self):
        return "%s/%d.rdf" % (self.dirname(), self.id)

    def load(self):
        path = self.path()
        if os.path.isfile(path):
            f = file(path)
            handler = RSShandler()
            parse(f, handler)
            self.item = handler.items[0]
            return True
        else:
            return False

    def sync(self):
        dirname = self.dirname()
        if not os.path.isdir(dirname):
            os.makedirs(dirname)
        rss = RSS()
        rss[self.item.link] = self.item
        f = file(self.path(), "w")
        buf = make_rss1(rss)
        f.write(buf)
        return buf

    def remove(self):
        path = self.path()
        dirname = self.dirname()
        os.remove(path)
        files = os.listdir(dirname)
        if len(files) == 0:
            os.rmdir(dirname)
        elif len(files) == 1:
            os.remove(os.path.join(dirname, files[0]))
            os.rmdir(dirname)


def error(msg="Something wrong"):
    print "Content-Type: text/xml; charset=UTF-8"
    print
    print '<?xml version="1.0" encoding="utf-8"?>'
    print '<error>%s</error>' % msg

def update_archives():
    rss = RSS(uri = config["archivesuri"])
    parent = config["link"]
    for i in os.listdir(config["xmldir"]):
        if re.search(r"^\d+-\d+", i):
            rss.append(link = "%s#a%s" % (parent, i),
                       title = i,
                       date = i)
    f = file(config["archives"], "w")
    f.write(make_rss1(rss))

def update_recent(rss_cache):
    rss = RSS(uri = config["rssuri"])
    month = []
    ids = []
    xmldir = config["xmldir"]
    for i in os.listdir(xmldir):
        if re.search(r"^\d+-\d+", i):
            month.append(i)
    month.sort()
    month.reverse()
    rsssize = config.getint("rsssize")
    for i in month:
        for j in os.listdir(os.path.join(xmldir, i)):
            found = re.search(r"^(\d+)", j)
            if found:
                ids.append(found.group(1))
        if len(ids) >= rsssize:
            break
    count = 0
    ids.sort()
    ids.reverse()
    for id in ids:
        if id in rss_cache:
            item = rss_cache[id]
        else:
            entry = Entry(int(id))
            entry.load()
            item = entry.item
        rss[item.link] = item
        rss_cache[item.link] = item
        count += 1
        if count >= rsssize:
            break
    f = file(config["recent"], "w")
    f.write(make_rss1(rss))

def update_monthly(month, rss_cache):
    rss = RSS(uri = urljoin(config["xmluri"], month + "/index.rdf"))
    dirname = os.path.join(config["xmldir"], month)
    if not os.path.isdir(dirname):
        return
    for i in os.listdir(dirname):
        found = re.search(r"^(\d+)", i)
        if found:
            id = found.group(1)
            if id in rss_cache:
                item = rss_cache[id]
            else:
                entry = Entry(int(id))
                entry.load()
                item = entry.item
            item.content = ""
            rss[item.link] = item
            rss_cache[item.link] = item

    f = file(os.path.join(dirname, "index.rdf"), "w")
    f.write(make_rss1(rss))

def write(id="", title="", body=""):
    title = title.strip()
    body = body.strip()
    try:
        if not id:
            id = int(time.time())
        else:
            id = int(id)
    except ValueError:
        error("Wrong entry ID.")
        return
    if len(title)+len(body) > config.getint("pagesize"):
        error("Too large entry.")
        return
    entry = Entry(id)
    item = entry.update(title=title, content=body)
    rss_cache = {}
    rss_cache[str(id)] = item
    if title or body:
        entry.sync()
    update_archives()
    update_recent(rss_cache)
    update_monthly(entry.month(), rss_cache)
    if title or body:
        print "Content-Type: text/xml; charset=UTF-8"
        print
        print '<?xml version="1.0" encoding="utf-8"?>'
        print '<ok>ok</ok>'
    else:
        error("Removed")


def main():
    try:
        os.environ["TZ"] = config["tz"]
        time.tzset()
        form = cgi.FieldStorage()
        id = form.getfirst("id", "")
        title = form.getfirst("title", "")
        passwd = form.getfirst("passwd", "")
        body = form.getfirst("body", "")
        if md5.new(passwd).hexdigest() == config["passwd"]:
            write(id, title, body)
        else:
            error("Wrong password.")

    except (KeyError, ValueError, IOError, OSError), e:
        error(e)

if __name__ == "__main__":
    main()
